import logging
from typing import Optional

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration


def _tag_trace_id(event: dict) -> None:
    if "request" not in event:
        return

    if not isinstance(event["request"], dict):
        return

    url = event["request"].get("url")
    if not isinstance(url, str):
        return

    if not url.endswith("/workflow/stream"):
        return

    body = event["request"].get("data")
    if not isinstance(body, dict):
        return

    execution_context = body.get("execution_context")
    if not isinstance(execution_context, dict):
        return

    trace_id = execution_context.get("trace_id")
    if not isinstance(trace_id, str):
        return

    if "tags" not in event:
        event["tags"] = {}

    event["tags"]["vellum_trace_id"] = trace_id


def _apply_custom_tags(event: dict, hint: dict) -> None:
    """
    Extracts 'sentry_tags' from logger exception's extra data and adds them to the event tags.

    Modifies the event dictionary in place.
    """
    record = hint.get("log_record")
    if not record:
        return

    sentry_tags = getattr(record, "sentry_tags", None)
    if not sentry_tags or not isinstance(sentry_tags, dict):
        return

    if "tags" not in event:
        event["tags"] = {}

    event["tags"].update(sentry_tags)


def before_send(event: dict, hint: dict) -> Optional[dict]:
    if "exc_info" in hint:
        _, _, _ = hint["exc_info"]

    _tag_trace_id(event)
    _apply_custom_tags(event, hint)

    return event


def init_sentry(*, sentry_log_level: Optional[int], dsn: str) -> None:
    sentry_logging = LoggingIntegration(
        level=sentry_log_level,  # Capture info and above as breadcrumbs
        event_level=logging.ERROR,  # Send errors as events
    )
    integrations = [sentry_logging]

    sentry_sdk.init(
        dsn=dsn,
        integrations=integrations,
        environment="production",
        traces_sample_rate=0.0,
        send_default_pii=True,
        before_send=before_send,  # type: ignore
        max_request_body_size="always",
        max_value_length=10_000,
    )



from unittest.mock import mock_open, patch

from workflow_server.config import MEMORY_LIMIT_MB
from workflow_server.utils.system_utils import (
    FORCE_GC_MEMORY_PERCENT,
    WARN_MEMORY_PERCENT,
    get_memory_in_use_mb,
    wait_for_available_process,
)


def test_get_memory_in_use_mb_success():
    # Test with 1GB of memory (1024MB)
    test_memory_bytes = "1073741824"
    with patch("builtins.open", mock_open(read_data=test_memory_bytes)):
        result = get_memory_in_use_mb()

        assert result == 1024.0


def test_get_memory_in_use_mb_empty_file():
    with patch("builtins.open", mock_open(read_data="")):
        result = get_memory_in_use_mb()

        assert result is None


def test_get_memory_in_use_mb_file_not_found():
    with patch("builtins.open", side_effect=FileNotFoundError()):
        result = get_memory_in_use_mb()

        assert result is None


def test_get_memory_in_use_mb_zero_memory():
    with patch("builtins.open", mock_open(read_data="0")):
        result = get_memory_in_use_mb()
        assert result == 0.0


@patch("workflow_server.utils.system_utils.time.sleep")
@patch("workflow_server.utils.system_utils.get_memory_in_use_mb")
@patch("workflow_server.utils.system_utils.get_active_process_count")
def test_wait_for_available_process_immediate_availability(mock_get_active_process_count, mock_get_memory, mock_sleep):
    # Mock memory usage below warning limit and process limit below
    mock_get_memory.return_value = MEMORY_LIMIT_MB * (WARN_MEMORY_PERCENT - 0.1)
    mock_get_active_process_count.return_value = 10

    result = wait_for_available_process()

    assert result is True

    # Should not sleep if immediately available
    mock_sleep.assert_not_called()


@patch("workflow_server.utils.system_utils.time.sleep")
@patch("workflow_server.utils.system_utils.get_memory_in_use_mb")
@patch("workflow_server.utils.system_utils.get_active_process_count")
def test_wait_for_available_process_becomes_available(mock_get_active_process_count, mock_get_memory, mock_sleep):
    # First two calls indicate high memory usage, third call shows available memory
    mock_get_memory.side_effect = [
        MEMORY_LIMIT_MB * (WARN_MEMORY_PERCENT + 0.1),
        MEMORY_LIMIT_MB * (WARN_MEMORY_PERCENT + 0.1),
        MEMORY_LIMIT_MB * (WARN_MEMORY_PERCENT - 0.1),
    ]
    mock_get_active_process_count.return_value = 10

    result = wait_for_available_process()

    assert result is True
    # Should sleep twice before becoming available
    assert mock_sleep.call_count == 2


@patch("workflow_server.utils.system_utils.time.sleep")
@patch("workflow_server.utils.system_utils.get_memory_in_use_mb")
@patch("workflow_server.utils.system_utils.get_active_process_count")
def test_wait_for_available_process_never_available(mock_get_active_process_count, mock_get_memory, mock_sleep):
    # Return false if process isn't available from high memory usage
    mock_get_memory.return_value = MEMORY_LIMIT_MB * (WARN_MEMORY_PERCENT + 0.1)
    mock_get_active_process_count.return_value = 13

    result = wait_for_available_process()
    assert result is False

    # Should sleep for each attempt
    assert mock_sleep.call_count == 3


@patch("workflow_server.utils.system_utils.time.sleep")
@patch("workflow_server.utils.system_utils.get_memory_in_use_mb")
@patch("workflow_server.utils.system_utils.get_active_process_count")
def test_wait_for_available_process_memory_none(mock_get_active_process_count, mock_get_memory, mock_sleep):
    # Test when memory reading fails that result is still true
    mock_get_memory.return_value = None
    mock_get_active_process_count.return_value = 10

    result = wait_for_available_process()
    assert result is True


@patch("workflow_server.utils.system_utils.time.sleep")
@patch("workflow_server.utils.system_utils.get_memory_in_use_mb")
@patch("workflow_server.utils.system_utils.get_active_process_count")
def test_wait_for_available_process_high_process_count_but_low_memory(
    mock_get_active_process_count, mock_get_memory, mock_sleep
):
    # Test when process count is high but memory is low
    mock_get_memory.return_value = MEMORY_LIMIT_MB * (FORCE_GC_MEMORY_PERCENT - 0.1)
    mock_get_active_process_count.return_value = 13

    result = wait_for_available_process()
    assert result is True



import json
import logging
import re
from unittest.mock import patch
from uuid import UUID, uuid4

from deepdiff import DeepDiff

from workflow_server.server import create_app


def test_version_route():
    flask_app = create_app()

    with flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")
        assert response.status_code == 200
        assert re.match(r"[0-9]*\.[0-9]*\.[0-9]*", response.json["sdk_version"])
        assert response.json["server_version"] == "local"


def test_version_route__with_single_node_file(tmp_path):
    # GIVEN a temporary custom_nodes directory with a test node
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    node_file = custom_nodes_dir / "test_node.py"
    node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class TestNode(BaseNode):
    \"""A test node for processing data.\"""
"""
    )

    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND we should find exactly one node
    nodes = response.json["nodes"]
    assert len(nodes) == 1

    # AND the node should have the correct metadata
    node = nodes[0]
    assert UUID(node["id"])
    assert node["module"] == ["vellum_custom_nodes", "test_node"]
    assert node["name"] == "TestNode"
    assert node["label"] == "Test Node"
    assert "A test node for processing data." in node["description"]
    assert node["exec_config"] == {
        "adornments": None,
        "attributes": [],
        "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
        "definition": {"module": ["vellum_custom_nodes", "test_node"], "name": "TestNode"},
        "display_data": {
            "comment": {"expanded": True, "value": "A test node for processing data."},
            "position": {"x": 0.0, "y": 0.0},
        },
        "id": "7a8b251d-f5ca-462a-b293-071d219460fb",
        "label": "Test Node",
        "outputs": [],
        "ports": [{"id": "a3a0eefd-45d0-4f13-8c58-a836a9f7f9ed", "name": "default", "type": "DEFAULT"}],
        "trigger": {"id": "a022e36c-9852-4772-9be3-3c6c147fd811", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "type": "GENERIC",
    }


def test_version_route__with_nodes_in_multiple_files(tmp_path):
    # GIVEN a temporary custom_nodes directory
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    # AND a first node file
    first_node_file = custom_nodes_dir / "first_node.py"
    first_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class SomeNode(BaseNode):
    \"""This is Some Node.\"""
"""
    )

    # AND a second node file
    second_node_file = custom_nodes_dir / "second_node.py"
    second_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class SomeOtherNode(BaseNode):
    \"""This is Some Other Node.\"""
"""
    )

    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND we should find both nodes
    nodes = response.json["nodes"]
    assert len(nodes) == 2

    # AND the first node should have correct metadata
    some_node = nodes[0]
    assert some_node["label"] == "Some Node"
    assert some_node["description"] == "This is Some Node."
    assert UUID(some_node["id"])
    assert some_node["module"] == ["vellum_custom_nodes", "first_node"]
    assert some_node["exec_config"] == {
        "adornments": None,
        "attributes": [],
        "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
        "definition": {"module": ["vellum_custom_nodes", "first_node"], "name": "SomeNode"},
        "display_data": {
            "comment": {"expanded": True, "value": "This is Some Node."},
            "position": {"x": 0.0, "y": 0.0},
        },
        "id": "1e559c2e-db82-41f0-9ceb-5e89b0c5a0a3",
        "label": "Some Node",
        "outputs": [],
        "ports": [{"id": "48e39e97-5fd4-471e-b4f2-51d3baf06456", "name": "default", "type": "DEFAULT"}],
        "trigger": {"id": "e3381fb7-61fc-4c46-ae8e-51fc463b6a59", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "type": "GENERIC",
    }

    # AND the second node should have correct metadata
    some_other_node = nodes[1]
    assert some_other_node["label"] == "Some Other Node"
    assert some_other_node["description"] == "This is Some Other Node."
    assert UUID(some_other_node["id"])
    assert some_other_node["module"] == ["vellum_custom_nodes", "second_node"]
    assert some_other_node["exec_config"] == {
        "adornments": None,
        "attributes": [],
        "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
        "definition": {"module": ["vellum_custom_nodes", "second_node"], "name": "SomeOtherNode"},
        "display_data": {
            "comment": {"expanded": True, "value": "This is Some Other Node."},
            "position": {"x": 0.0, "y": 0.0},
        },
        "id": "7aee541b-b245-4c8a-9137-3e4631d5100c",
        "label": "Some Other Node",
        "outputs": [],
        "ports": [{"id": "fb66b46a-d970-4bc9-83ea-70c154c57ddd", "name": "default", "type": "DEFAULT"}],
        "trigger": {"id": "13fa2714-20b3-4bc3-ab79-621a188e3bfa", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "type": "GENERIC",
    }


def test_version_route__no_custom_nodes_dir(tmp_path):
    # GIVEN a Flask application and an empty temp directory
    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND the nodes list should be empty
    assert response.json["nodes"] == []


def test_version_route__with_multiple_nodes_in_file(tmp_path):
    # Create a temporary custom_nodes directory with multiple nodes in one file
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    # Create a test node file with multiple nodes
    node_file = custom_nodes_dir / "multiple_nodes.py"
    node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class ProcessingNode(BaseNode):
    \"""Processes input data.\"""

class TransformationNode(BaseNode):
    \"""Transforms data format.\"""

# This class should not be discovered
class HelperClass:
    pass
"""
    )

    flask_app = create_app()

    # Mock the current working directory to point to our temp directory
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

        assert response.status_code == 200
        nodes = response.json["nodes"]
        assert len(nodes) == 2

        # Nodes should be discovered regardless of their order in the file
        node_names = {node["name"] for node in nodes}
        assert node_names == {"ProcessingNode", "TransformationNode"}

        # Find and assert each node individually
        processing_node = next(node for node in nodes if node["name"] == "ProcessingNode")
        assert processing_node["exec_config"] == {
            "adornments": None,
            "attributes": [],
            "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
            "definition": {"module": ["vellum_custom_nodes", "multiple_nodes"], "name": "ProcessingNode"},
            "display_data": {
                "comment": {"expanded": True, "value": "Processes input data."},
                "position": {"x": 0.0, "y": 0.0},
            },
            "id": "f92c09f0-0434-46cb-829d-a73f801d6343",
            "label": "Processing Node",
            "outputs": [],
            "ports": [{"id": "abaa2984-b312-4491-b069-e689759f72c8", "name": "default", "type": "DEFAULT"}],
            "trigger": {"id": "35378c2b-f089-44af-ac37-efe4ea42c817", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "type": "GENERIC",
        }

        transformation_node = next(node for node in nodes if node["name"] == "TransformationNode")
        assert transformation_node["exec_config"] == {
            "adornments": None,
            "attributes": [],
            "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
            "definition": {"module": ["vellum_custom_nodes", "multiple_nodes"], "name": "TransformationNode"},
            "display_data": {
                "comment": {"expanded": True, "value": "Transforms data format."},
                "position": {"x": 0.0, "y": 0.0},
            },
            "id": "09ca32f7-c8f2-4469-97e5-1f288f85127a",
            "label": "Transformation Node",
            "outputs": [],
            "ports": [{"id": "88778117-fbfc-4b44-964b-5a4994aa2f24", "name": "default", "type": "DEFAULT"}],
            "trigger": {"id": "5d096263-7fbf-490a-83b7-e441852b5fb6", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "type": "GENERIC",
        }


def test_version_route__with_invalid_node_file(tmp_path, caplog):
    caplog.set_level(logging.WARNING)

    # GIVEN a temporary custom_nodes directory
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    # AND a valid node file
    valid_node_file = custom_nodes_dir / "valid_node.py"
    valid_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class SomeNode(BaseNode):
    \"\"\"This is Some Node.\"\"\"
"""
    )

    # AND an invalid node file with syntax error of missing colon in the class
    invalid_node_file = custom_nodes_dir / "invalid_node.py"
    invalid_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class BrokenNode(BaseNode)
    \"\"\"This node has a syntax error.\"\"\"
"""
    )

    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND we should find only the valid node
    nodes = response.json["nodes"]
    assert len(nodes) == 1

    # AND the valid node should have correct metadata
    valid_node = nodes[0]
    assert valid_node["label"] == "Some Node"
    assert valid_node["description"] == "This is Some Node."
    assert UUID(valid_node["id"])
    assert valid_node["module"] == ["vellum_custom_nodes", "valid_node"]
    assert valid_node["exec_config"] == {
        "adornments": None,
        "attributes": [],
        "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
        "definition": {"module": ["vellum_custom_nodes", "valid_node"], "name": "SomeNode"},
        "display_data": {
            "comment": {"expanded": True, "value": "This is Some Node."},
            "position": {"x": 0.0, "y": 0.0},
        },
        "id": "1e559c2e-db82-41f0-9ceb-5e89b0c5a0a3",
        "label": "Some Node",
        "outputs": [],
        "ports": [{"id": "48e39e97-5fd4-471e-b4f2-51d3baf06456", "name": "default", "type": "DEFAULT"}],
        "trigger": {"id": "e3381fb7-61fc-4c46-ae8e-51fc463b6a59", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "type": "GENERIC",
    }

    # AND the error should be logged with full traceback
    assert len(caplog.records) > 0
    error_message = caplog.records[0].message
    assert "Failed to load node from module invalid_node" in error_message
    assert "invalid_node.py, line 4" in error_message


def test_version_route__with_attributes(tmp_path):
    # GIVEN a temporary custom_nodes directory
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    # AND an addition node file
    node_file = custom_nodes_dir / "addition_node.py"
    node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode


class MyAdditionNode(BaseNode):
    \"\"\"Custom node that performs simple addition.\"\"\"
    arg1: int
    arg2: int

    class Outputs(BaseNode.Outputs):
        result: int

    def run(self) -> BaseNode.Outputs:
        result = self.arg1 + self.arg2
        return self.Outputs(result=result)
"""
    )

    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND we should find the addition node
    nodes = response.json["nodes"]
    assert len(nodes) == 1

    # AND the node should have the correct metadata
    node = nodes[0]
    assert node["label"] == "My Addition Node"
    assert node["description"] == "Custom node that performs simple addition."
    assert UUID(node["id"])
    assert node["module"] == ["vellum_custom_nodes", "addition_node"]
    assert node["name"] == "MyAdditionNode"
    assert node["exec_config"] == {
        "adornments": None,
        "attributes": [
            {
                "id": "aed3bcbb-d243-4a77-bb5e-409e9a28e868",
                "name": "arg1",
                "value": None,
            },
            {
                "id": "9225d225-a41b-4642-8964-f28f58dcf4bf",
                "name": "arg2",
                "value": None,
            },
        ],
        "base": {"module": ["vellum", "workflows", "nodes", "bases", "base"], "name": "BaseNode"},
        "definition": {"module": ["vellum_custom_nodes", "addition_node"], "name": "MyAdditionNode"},
        "display_data": {
            "comment": {"expanded": True, "value": "Custom node that performs simple addition."},
            "position": {"x": 0.0, "y": 0.0},
        },
        "id": "195cd69d-3d2d-41e4-a432-16c433cb8d34",
        "label": "My Addition Node",
        "outputs": [{"id": "3d8e40cb-2aa8-44bd-ae6a-708a9fbc4779", "name": "result", "type": "NUMBER", "value": None}],
        "ports": [{"id": "9a9e4ef6-febf-4093-a515-217bbb1373db", "name": "default", "type": "DEFAULT"}],
        "trigger": {"id": "a5298668-d808-4a45-a62e-790943948e8a", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "type": "GENERIC",
        "should_file_merge": True,
    }


def test_serialize_route__with_no_files():
    # GIVEN a Flask application
    flask_app = create_app()

    # WHEN we make a request with no files
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/serialize", json={"files": {}})

    # THEN we should get a bad request response
    assert response.status_code == 400

    # AND the response should contain an error message
    assert "detail" in response.json
    assert "No files received" in response.json["detail"]


def test_serialize_route__with_invalid_python_syntax():
    # GIVEN a Flask application
    flask_app = create_app()

    # AND a file with invalid Python syntax
    invalid_content = """
from vellum.workflows.nodes import BaseNode

class BrokenNode(BaseNode)  # Missing colon
    \"\"\"This node has a syntax error.\"\"\"
"""

    # WHEN we make a request to the serialize route
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/serialize", json={"files": {"broken_node.py": invalid_content}})

    # THEN we should get a 400 response
    assert response.status_code == 400

    # AND the response should contain an error message
    assert "detail" in response.json
    assert "Serialization failed" in response.json["detail"]


def test_serialize_route__with__workflow():
    # GIVEN a Flask application
    flask_app = create_app()

    # AND a complete workflow with multiple files
    workflow_files = {
        "__init__.py": "# flake8: noqa: F401, F403\n\n",
        "inputs.py": "from typing import Any, Optional\n\nfrom vellum.workflows.inputs import BaseInputs\n\n\nclass Inputs(BaseInputs):\n    text: str\n    var_1: Optional[Any]\n",  # noqa: E501
        "workflow.py": "from vellum.workflows import BaseWorkflow\nfrom vellum.workflows.state import BaseState\n\nfrom .inputs import Inputs\nfrom .nodes.final_output import FinalOutput\nfrom .nodes.templating_node import TemplatingNode\n\n\nclass Workflow(BaseWorkflow[Inputs, BaseState]):\n    graph = TemplatingNode >> FinalOutput\n\n    class Outputs(BaseWorkflow.Outputs):\n        final_output = FinalOutput.Outputs.value\n",  # noqa: E501
        "nodes/__init__.py": 'from .final_output import FinalOutput\nfrom .templating_node import TemplatingNode\n\n__all__ = [\n    "FinalOutput",\n    "TemplatingNode",\n]\n',  # noqa: E501
        "nodes/templating_node.py": 'from vellum.workflows.nodes.displayable import TemplatingNode as BaseTemplatingNode\nfrom vellum.workflows.state import BaseState\n\nfrom ..inputs import Inputs\n\n\nclass TemplatingNode(BaseTemplatingNode[BaseState, str]):\n    template = """{{ text }}"""\n    inputs = {\n        "text": Inputs.text,\n    }\n',  # noqa: E501
        "nodes/final_output.py": "from vellum.workflows.nodes.displayable import FinalOutputNode\nfrom vellum.workflows.state import BaseState\n\nfrom .templating_node import TemplatingNode\n\n\nclass FinalOutput(FinalOutputNode[BaseState, str]):\n    class Outputs(FinalOutputNode.Outputs):\n        value = TemplatingNode.Outputs.result\n",  # noqa: E501
    }

    # WHEN we make a request to the serialize route
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/serialize", json={"files": workflow_files})

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND the response should contain the full WorkflowSerializationResult
    assert "exec_config" in response.json
    assert "errors" in response.json
    assert "dataset" in response.json or response.json.get("dataset") is None
    exec_config = response.json["exec_config"]

    # AND the exec_config should have workflow_raw_data
    assert "workflow_raw_data" in exec_config
    nodes = exec_config["workflow_raw_data"]["nodes"]

    # AND we should find the workflow nodes
    node_labels = {node["data"]["label"] for node in nodes}
    expected_nodes = {"Templating Node", "Final Output", "Entrypoint Node"}

    # AND at least some of the expected nodes should be present
    assert not DeepDiff(node_labels, expected_nodes, ignore_order=True)


def test_serialize_route__with_workspace_api_key():
    """
    Tests that the serialize route accepts workspace_api_key and passes it through serialization.
    """
    # GIVEN a Flask application
    flask_app = create_app()

    # AND a complete workflow with multiple files
    workflow_files = {
        "__init__.py": "# flake8: noqa: F401, F403\n\n",
        "inputs.py": "from typing import Any, Optional\n\nfrom vellum.workflows.inputs import BaseInputs\n\n\nclass Inputs(BaseInputs):\n    text: str\n    var_1: Optional[Any]\n",  # noqa: E501
        "workflow.py": "from vellum.workflows import BaseWorkflow\nfrom vellum.workflows.state import BaseState\n\nfrom .inputs import Inputs\nfrom .nodes.final_output import FinalOutput\nfrom .nodes.templating_node import TemplatingNode\n\n\nclass Workflow(BaseWorkflow[Inputs, BaseState]):\n    graph = TemplatingNode >> FinalOutput\n\n    class Outputs(BaseWorkflow.Outputs):\n        final_output = FinalOutput.Outputs.value\n",  # noqa: E501
        "nodes/__init__.py": 'from .final_output import FinalOutput\nfrom .templating_node import TemplatingNode\n\n__all__ = [\n    "FinalOutput",\n    "TemplatingNode",\n]\n',  # noqa: E501
        "nodes/templating_node.py": 'from vellum.workflows.nodes.displayable import TemplatingNode as BaseTemplatingNode\nfrom vellum.workflows.state import BaseState\n\nfrom ..inputs import Inputs\n\n\nclass TemplatingNode(BaseTemplatingNode[BaseState, str]):\n    template = """{{ text }}"""\n    inputs = {\n        "text": Inputs.text,\n    }\n',  # noqa: E501
        "nodes/final_output.py": "from vellum.workflows.nodes.displayable import FinalOutputNode\nfrom vellum.workflows.state import BaseState\n\nfrom .templating_node import TemplatingNode\n\n\nclass FinalOutput(FinalOutputNode[BaseState, str]):\n    class Outputs(FinalOutputNode.Outputs):\n        value = TemplatingNode.Outputs.result\n",  # noqa: E501
    }

    # WHEN we make a request to the serialize route with workspace_api_key
    with flask_app.test_client() as test_client:
        response = test_client.post(
            "/workflow/serialize", json={"files": workflow_files, "workspace_api_key": "test_workspace_key"}
        )

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND the response should contain the full WorkflowSerializationResult
    assert "exec_config" in response.json
    assert "errors" in response.json


def test_serialize_route__with_invalid_workspace_api_key():
    """
    Tests that the serialize route handles invalid workspace_api_key gracefully.
    """
    # GIVEN a Flask application
    flask_app = create_app()

    workflow_files = {
        "__init__.py": "",
        "workflow.py": (
            "from vellum.workflows import BaseWorkflow\n\n"
            "class Workflow(BaseWorkflow):\n"
            "    class Outputs(BaseWorkflow.Outputs):\n"
            "        foo = 'hello'\n"
        ),
    }

    # WHEN we make a request with an invalid workspace_api_key
    with flask_app.test_client() as test_client:
        response = test_client.post(
            "/workflow/serialize", json={"files": workflow_files, "workspace_api_key": ""}  # Invalid empty key
        )

    # THEN we should still get a successful response (graceful degradation)
    assert response.status_code == 200

    # AND the response should contain the serialization result
    assert "exec_config" in response.json


def test_stream_node_route__with_node_id():
    """
    Tests that the stream-node endpoint works with node_id.
    """
    # GIVEN a valid request body with node_id
    node_id = uuid4()
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "node_id": str(node_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "from .display import *",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from .nodes.test_node import TestNode

class Workflow(BaseWorkflow):
    graph = TestNode

    class Outputs(BaseWorkflow.Outputs):
        result = TestNode.Outputs.value
""",
            "nodes/__init__.py": "from .test_node import TestNode\n__all__ = ['TestNode']",
            "nodes/test_node.py": """\
from vellum.workflows.nodes import BaseNode

class TestNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value = "test_result"
""",
            "display/__init__.py": "from .nodes import *\nfrom .workflow import *",
            "display/workflow.py": """\
from vellum_ee.workflows.display.workflows import BaseWorkflowDisplay
""",
            "display/nodes/__init__.py": "from .test_node import TestNodeDisplay\n__all__ = ['TestNodeDisplay']",
            "display/nodes/test_node.py": f"""\
from uuid import UUID
from vellum_ee.workflows.display.nodes import BaseNodeDisplay
from ...nodes.test_node import TestNode

class TestNodeDisplay(BaseNodeDisplay[TestNode]):
    node_id = UUID("{node_id}")
""",
        },
    }

    flask_app = create_app()

    # WHEN we call the stream-node route
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream-node", json=request_body)

    # THEN we get a 200 response
    assert response.status_code == 200

    events = [json.loads(line) for line in response.data.decode().strip().split("\n") if line.strip()]
    assert len(events) == 4
    assert events[0]["name"] == "vembda.execution.initiated"
    assert events[1]["name"] == "node.execution.initiated"
    assert events[2]["name"] == "node.execution.fulfilled"
    assert events[3]["name"] == "vembda.execution.fulfilled"


def test_stream_node_route__with_node_module_and_name_backward_compatibility():
    """
    Tests that the stream-node endpoint still works with node_module and node_name for backward compatibility.
    """
    # GIVEN a valid request body with node_module and node_name (old format)
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "node_module": "nodes.test_node",
        "node_name": "TestNode",
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from .nodes.test_node import TestNode

class Workflow(BaseWorkflow):
    graph = TestNode
""",
            "nodes/__init__.py": "from .test_node import TestNode",
            "nodes/test_node.py": """\
from vellum.workflows.nodes import BaseNode

class TestNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value = "test_result"
""",
        },
    }

    flask_app = create_app()

    # WHEN we call the stream-node route
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream-node", json=request_body)

    # THEN we get a 200 response
    assert response.status_code == 200

    events = [json.loads(line) for line in response.data.decode().strip().split("\n") if line.strip()]
    assert events[0]["name"] == "vembda.execution.initiated"
    assert events[1]["name"] == "node.execution.initiated", json.dumps(events[1]["body"])
    assert events[2]["name"] == "node.execution.fulfilled"
    assert events[3]["name"] == "vembda.execution.fulfilled"
    assert len(events) == 4


def test_stream_node_route__missing_node_info_validation():
    """
    Tests that the stream-node endpoint returns validation error when neither
    node_id nor node_module/node_name are provided.
    """
    # GIVEN a request body missing node identification
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {"__init__.py": "", "workflow.py": ""},
    }

    flask_app = create_app()

    # WHEN we call the stream-node route
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream-node", json=request_body)

    # THEN we get a 400 response
    assert response.status_code == 400

    # AND we get a validation error message
    assert "Either node_id or both node_module and node_name must be provided" in response.get_json()["detail"]


def test_stream_node_route__invalid_node_id():
    """
    Tests that the stream-node endpoint returns 404 for invalid node_id.
    """
    # GIVEN a request body with invalid node_id
    invalid_node_id = uuid4()
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "node_id": str(invalid_node_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow

class Workflow(BaseWorkflow):
    pass
""",
        },
    }

    flask_app = create_app()

    # WHEN we call the stream-node route
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream-node", json=request_body)

    # THEN we get a 200 response
    # TODO: In the future, we would want this to return a 4xx response by returning the workflow
    # instance and the node definition that we want to run as part of request deserialization.
    assert response.status_code == 200, response.text

    # AND we get an appropriate error message
    # TODO: In a future where we are returning 4xx responses, we assert the following data:
    # response_data = response.get_json()
    # assert "Node with ID" in response_data["detail"]
    # assert "not found" in response_data["detail"]
    events = [json.loads(line) for line in response.data.decode().strip().split("\n") if line.strip()]
    assert events[0]["name"] == "vembda.execution.initiated"
    assert events[1]["name"] == "vembda.execution.fulfilled"
    assert len(events) == 2


def test_serialize_route__with_invalid_nested_set_graph():
    """
    Tests that a workflow with an invalid nested set graph structure raises a clear user-facing exception.
    """
    # GIVEN a Flask application
    flask_app = create_app()

    invalid_workflow_content = """
from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes import BaseNode

class TestNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value = "test"

class InvalidWorkflow(BaseWorkflow):
    graph = {TestNode, {TestNode}}

    class Outputs(BaseWorkflow.Outputs):
        result = TestNode.Outputs.value
"""

    workflow_files = {
        "__init__.py": "",
        "workflow.py": invalid_workflow_content,
    }

    # WHEN we make a request to the serialize route
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/serialize", json={"files": workflow_files})

    # THEN we should get a 400 response
    assert response.status_code == 400

    # AND the response should contain a user-friendly error message
    assert "detail" in response.json
    error_detail = response.json["detail"]
    assert "Serialization failed" in error_detail
    assert "Invalid graph structure detected" in error_detail
    assert "contact Vellum support" in error_detail



import os

from dotenv import load_dotenv

load_dotenv()

VEMBDA_PUBLIC_KEY = """
-----BEGIN PUBLIC KEY-----
MIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA3VrSo+PIYLsKaLwp6Kuw
TKzWHaAfJuoVaZtcc0WCVV6nS3wuyTtzHSc81gNT4AotIMgYExu80pG64MwZ2f57
iQPeMXSMQwcp49PRTwwAYXrb/XFhra56mwPOmNLPFHz4hChvJvAx8R4h+otUvkEZ
BRGgOICV5eMsELpvEz1Dn/pnw42UcNFI8kEAZJfVaj8izsGlYDK55QOZB+D/hETR
s7V0g9r9zjYwxmwNAAlRkQtJ4BzbrztqJt5uUVGEAmORgGdCsRU81Fd42GyHZxVY
WaAoPuqFj/Hd9PPvKuSZk/HBsxNBbyEp1zFoi8MxaVBgCpcWVEQWAsRo+ak9v6oN
yxxfZx8zshV0z2c79OpVre8rd+3dJviCIjTgClOe+ior0bpUSvd3q3gj5/ZLWo2v
uRJxsH4EXcJndcp3wbC4y98N61towHwzKdDIr1y3vmO2kh+m7MGg4mqMZs652Y4r
groREv+ohDej8CzU9CWSx5dE2leA1iIzLvKlbpkCIaQ4oFBUh9DlIZuCRGufgXBU
Tav8XWrbhG9nKJ+Gb3Jsol2YmDn+cdDIT8gs+dpe2/UyXH7YE0Oz0X3BVek17nej
nnTwu1M91VsPGq0tTcSi/AY53EO51YZr77Ln8ue694l2emyJDyj041bAmPAQOb26
czy4Vhr2WxlC6p9ekb3PY/8CAwEAAQ==
-----END PUBLIC KEY-----
"""

NAMESPACE = os.getenv("NAMESPACE")
IS_VPC = os.getenv("IS_VPC", "").lower() == "true"
MEMORY_LIMIT_MB = int(os.getenv("MEMORY_LIMIT_MB", "2048"))
PORT = os.getenv("PORT", "8000")
VELLUM_API_URL_HOST = os.getenv("VELLUM_API_URL_HOST", "localhost")
VELLUM_API_URL_PORT = os.getenv("VELLUM_API_URL_PORT", 8000)

# Set VELLUM_API_URL for SDK when using local Vellum API
if os.getenv("USE_LOCAL_VELLUM_API") == "true" and "VELLUM_API_URL" not in os.environ:
    os.environ["VELLUM_API_URL"] = "http://{}:{}".format(VELLUM_API_URL_HOST, VELLUM_API_URL_PORT)

CONCURRENCY = int(os.getenv("CONCURRENCY", "8"))
CONTAINER_IMAGE = os.getenv("CONTAINER_IMAGE", "python-workflow-runtime:latest")
ENABLE_PROCESS_WRAPPER = os.getenv("ENABLE_PROCESS_WRAPPER", "true").lower() == "true"

# This controls if we should just load a module already available in the py path for running
# a workflow. Currently used for running agent builder end to end locally
LOCAL_WORKFLOW_MODULE = os.getenv("LOCAL_WORKFLOW_MODULE")
# The deployment name to match against when using local mode so you can still run your normal workflow
LOCAL_DEPLOYMENT = os.getenv("LOCAL_DEPLOYMENT")


def is_development() -> bool:
    return os.getenv("FLASK_ENV", "local") == "local"



from datetime import datetime
import json
import logging
import os
from threading import Event as ThreadingEvent
from uuid import uuid4
from typing import Optional

from workflow_server.core.events import VembdaExecutionInitiatedBody, VembdaExecutionInitiatedEvent
from workflow_server.core.executor import stream_workflow
from workflow_server.core.utils import serialize_vembda_rejected_event
from workflow_server.core.workflow_executor_context import WorkflowExecutorContext
from workflow_server.utils.utils import get_version

logger = logging.getLogger(__name__)

_EVENT_LINE = "--event--"


def run_code_exec_stream() -> None:
    context: Optional[WorkflowExecutorContext] = None

    try:
        input_raw = ""
        while "--vellum-input-stop--" not in input_raw:
            # os/read they come in chunks of idk length, not as lines
            input_raw += os.read(0, 100_000_000).decode("utf-8")

        split_input = input_raw.split("\n--vellum-input-stop--\n")
        input_json = split_input[0]

        input_data = json.loads(input_json)
        context = WorkflowExecutorContext.model_validate(input_data)

        print("--vellum-output-start--")  # noqa: T201

        initiated_event = VembdaExecutionInitiatedEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=context.trace_id,
            span_id=context.execution_id,
            body=VembdaExecutionInitiatedBody.model_validate(get_version()),
            parent=None,
        ).model_dump_json()

        print(f"{_EVENT_LINE}{initiated_event}")  # noqa: T201

        stream_iterator, span_id = stream_workflow(
            context,
            disable_redirect=True,
            # Timeouts are handled at the code exec level right now so just passing in an unused threading event
            timeout_signal=ThreadingEvent(),
            cancel_signal=ThreadingEvent(),
        )
        for line in stream_iterator:
            print(f"{_EVENT_LINE}{json.dumps(line)}")  # noqa: T201
    except Exception as e:
        logger.exception(e)

        event = serialize_vembda_rejected_event(context, "Internal Server Error")
        print(f"{_EVENT_LINE}{event}")  # noqa: T201



from datetime import datetime
import importlib
from io import StringIO
import json
import logging
from multiprocessing import Process, Queue
import os
import random
import string
import sys
from threading import Event as ThreadingEvent
import time
from traceback import format_exc
from uuid import UUID, uuid4
from typing import Any, Callable, Generator, Iterator, Optional, Tuple, Type

from vellum_ee.workflows.display.utils.events import event_enricher
from vellum_ee.workflows.server.virtual_file_loader import VirtualFileFinder

from vellum.workflows import BaseWorkflow
from vellum.workflows.context import execution_context
from vellum.workflows.emitters.base import BaseWorkflowEmitter
from vellum.workflows.emitters.vellum_emitter import VellumEmitter
from vellum.workflows.events.exception_handling import stream_initialization_exception
from vellum.workflows.events.types import BaseEvent
from vellum.workflows.events.workflow import WorkflowEvent
from vellum.workflows.exceptions import WorkflowInitializationException
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes import BaseNode
from vellum.workflows.nodes.mocks import MockNodeExecution
from vellum.workflows.resolvers.base import BaseWorkflowResolver
from vellum.workflows.resolvers.resolver import VellumResolver
from vellum.workflows.state.context import WorkflowContext
from vellum.workflows.state.store import EmptyStore
from vellum.workflows.types import CancelSignal
from vellum.workflows.workflows.event_filters import workflow_sandbox_event_filter
from workflow_server.config import LOCAL_DEPLOYMENT, LOCAL_WORKFLOW_MODULE
from workflow_server.core.cancel_workflow import CancelWorkflowWatcherThread
from workflow_server.core.events import (
    SPAN_ID_EVENT,
    STREAM_FINISHED_EVENT,
    VembdaExecutionFulfilledBody,
    VembdaExecutionFulfilledEvent,
)
from workflow_server.core.utils import (
    create_vembda_rejected_event,
    is_events_emitting_enabled,
    serialize_vembda_rejected_event,
)
from workflow_server.core.workflow_executor_context import (
    BaseExecutorContext,
    NodeExecutorContext,
    WorkflowExecutorContext,
)
from workflow_server.utils.log_proxy import redirect_log
from workflow_server.utils.system_utils import get_memory_in_use_mb
from workflow_server.utils.utils import get_version

logger = logging.getLogger(__name__)


def stream_node_process_timeout(
    executor_context: NodeExecutorContext,
    queue: Queue,
) -> Process:
    node_process = Process(
        target=_stream_node_wrapper,
        args=(executor_context, queue),
    )
    node_process.start()

    if node_process.exitcode is not None:
        queue.put(create_vembda_rejected_event(executor_context, "Internal Server Error", timed_out=True))

    return node_process


def _stream_node_wrapper(executor_context: NodeExecutorContext, queue: Queue) -> None:
    try:
        for event in stream_node(executor_context=executor_context):
            queue.put(event)
    except WorkflowInitializationException as e:
        queue.put(create_vembda_rejected_event(executor_context, e.message))
    except Exception as e:
        logger.exception(e)
        queue.put(create_vembda_rejected_event(executor_context, "Internal Server Error"))


def _stream_workflow_wrapper(
    executor_context: WorkflowExecutorContext,
    queue: Queue,
    cancel_signal: CancelSignal,
    timeout_signal: CancelSignal,
) -> None:
    span_id_emitted = False
    try:
        stream_iterator, span_id = stream_workflow(
            executor_context=executor_context,
            cancel_signal=cancel_signal,
            timeout_signal=timeout_signal,
        )

        queue.put(f"{SPAN_ID_EVENT}:{span_id}")
        span_id_emitted = True

        for event in stream_iterator:
            queue.put(json.dumps(event))

    except Exception as e:
        if not span_id_emitted:
            queue.put(f"{SPAN_ID_EVENT}:{uuid4()}")

        logger.exception(e)
        queue.put(serialize_vembda_rejected_event(executor_context, "Internal Server Error"))

    queue.put(STREAM_FINISHED_EVENT)

    exit(0)


def stream_workflow_process_timeout(
    executor_context: WorkflowExecutorContext,
    queue: Queue,
    cancel_signal: CancelSignal,
    timeout_signal: CancelSignal,
) -> Process:
    workflow_process = Process(
        target=_stream_workflow_wrapper,
        args=(
            executor_context,
            queue,
            cancel_signal,
            timeout_signal,
        ),
    )
    workflow_process.start()

    if workflow_process.exitcode is not None:
        queue.put(create_vembda_rejected_event(executor_context, "Internal Server Error", timed_out=True))

    return workflow_process


def stream_workflow(
    executor_context: WorkflowExecutorContext,
    timeout_signal: CancelSignal,
    cancel_signal: CancelSignal,
    disable_redirect: bool = True,
) -> tuple[Iterator[dict], UUID]:
    cancel_watcher_kill_switch = ThreadingEvent()
    try:
        workflow, namespace = _create_workflow(executor_context)
        workflow_inputs = _get_workflow_inputs(executor_context, workflow.__class__)

        workflow_state = (
            workflow.deserialize_state(
                executor_context.state,
                workflow_inputs=workflow_inputs or BaseInputs(),
            )
            if executor_context.state
            else None
        )
        node_output_mocks = MockNodeExecution.validate_all(
            executor_context.node_output_mocks,
            workflow.__class__,
        )

        cancel_signal = cancel_signal or ThreadingEvent()

        stream = workflow.stream(
            inputs=workflow_inputs,
            state=workflow_state,
            node_output_mocks=node_output_mocks,
            event_filter=workflow_sandbox_event_filter,
            cancel_signal=cancel_signal,
            entrypoint_nodes=[executor_context.node_id] if executor_context.node_id else None,
            previous_execution_id=executor_context.previous_execution_id,
            timeout=executor_context.timeout,
        )
    except WorkflowInitializationException as e:
        cancel_watcher_kill_switch.set()

        with execution_context(
            parent_context=executor_context.execution_context.parent_context,
            trace_id=executor_context.execution_context.trace_id,
        ):
            initialization_exception_stream = stream_initialization_exception(e)

        def _stream_generator() -> Generator[dict[str, Any], Any, None]:
            for event in initialization_exception_stream:
                yield _dump_event(
                    event=event,
                    executor_context=executor_context,
                )

        return (
            _call_stream(
                executor_context=executor_context,
                stream_generator=_stream_generator,
                disable_redirect=disable_redirect,
                timeout_signal=timeout_signal,
            ),
            initialization_exception_stream.span_id,
        )
    except Exception:
        cancel_watcher_kill_switch.set()
        logger.exception("Failed to generate Workflow Stream")
        raise

    cancel_watcher = CancelWorkflowWatcherThread(
        kill_switch=cancel_watcher_kill_switch,
        execution_id=stream.span_id,
        timeout_seconds=executor_context.timeout,
        vembda_public_url=executor_context.vembda_public_url,
        cancel_signal=cancel_signal,
    )

    try:
        if executor_context.vembda_public_url:
            cancel_watcher.start()
    except Exception:
        logger.exception("Failed to start cancel watcher")

    def call_workflow() -> Generator[dict[str, Any], Any, None]:
        try:
            first = True
            for event in stream:
                if first:
                    executor_context.stream_start_time = time.time_ns()
                    first = False

                yield _dump_event(
                    event=event,
                    executor_context=executor_context,
                )
        except Exception as e:
            logger.exception("Failed to generate event from Workflow Stream")
            raise e
        finally:
            cancel_watcher_kill_switch.set()

        workflow.join()

    return (
        _call_stream(
            executor_context=executor_context,
            stream_generator=call_workflow,
            disable_redirect=disable_redirect,
            timeout_signal=timeout_signal,
        ),
        stream.span_id,
    )


def stream_node(
    executor_context: NodeExecutorContext,
    disable_redirect: bool = True,
) -> Iterator[dict]:
    workflow, namespace = _create_workflow(executor_context)
    Node: Optional[Type[BaseNode]] = None

    for workflow_node in workflow.get_nodes():
        if executor_context.node_id and workflow_node.__id__ == executor_context.node_id:
            Node = workflow_node
            break
        elif (
            executor_context.node_module
            and executor_context.node_name
            and workflow_node.__name__ == executor_context.node_name
            and workflow_node.__module__ == f"{namespace}.{executor_context.node_module}"
        ):
            Node = workflow_node
            break

    if not Node:
        identifier = executor_context.node_id or f"{executor_context.node_module}.{executor_context.node_name}"
        raise WorkflowInitializationException(
            message=f"Node '{identifier}' not found in workflow",
            workflow_definition=workflow.__class__,
        )

    def call_node() -> Generator[dict[str, Any], Any, None]:
        executor_context.stream_start_time = time.time_ns()

        for event in workflow.run_node(Node, inputs=executor_context.inputs):  # type: ignore[arg-type]
            yield event.model_dump(mode="json")

    return _call_stream(
        executor_context=executor_context,
        stream_generator=call_node,
        disable_redirect=disable_redirect,
        timeout_signal=ThreadingEvent(),
    )


def _call_stream(
    executor_context: BaseExecutorContext,
    stream_generator: Callable[[], Generator[dict[str, Any], Any, None]],
    timeout_signal: CancelSignal,
    disable_redirect: bool = True,
) -> Iterator[dict]:
    log_redirect: Optional[StringIO] = None

    if not disable_redirect:
        log_redirect = redirect_log()

    try:
        yield from stream_generator()

        vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=executor_context.trace_id,
            span_id=executor_context.execution_id,
            body=VembdaExecutionFulfilledBody(
                exit_code=0,
                log=log_redirect.getvalue() if log_redirect else "",
                stderr="",
                container_overhead_latency=executor_context.container_overhead_latency,
                timed_out=timeout_signal.is_set(),
            ),
            parent=None,
        )
        yield vembda_fulfilled_event.model_dump(mode="json")

    except Exception:
        vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=executor_context.trace_id,
            span_id=executor_context.execution_id,
            body=VembdaExecutionFulfilledBody(
                exit_code=-1,
                log=log_redirect.getvalue() if log_redirect else "",
                stderr=format_exc(),
                container_overhead_latency=executor_context.container_overhead_latency,
            ),
            parent=None,
        )
        yield vembda_fulfilled_event.model_dump(mode="json")


def _create_workflow(executor_context: BaseExecutorContext) -> Tuple[BaseWorkflow, str]:
    namespace = _get_file_namespace(executor_context)
    if namespace != LOCAL_WORKFLOW_MODULE:
        sys.meta_path.append(VirtualFileFinder(executor_context.files, namespace))

    workflow_context = _create_workflow_context(executor_context)
    Workflow = BaseWorkflow.load_from_module(namespace)
    VembdaExecutionFulfilledEvent.model_rebuild(
        # Not sure why this is needed, but it is required for the VembdaExecutionFulfilledEvent to be
        # properly rebuilt with the recursive types.
        # use flag here to determine which emitter to use
        _types_namespace={
            "BaseWorkflow": BaseWorkflow,
            "BaseNode": BaseNode,
        },
    )

    # Determine whether to enable the Vellum Emitter for event publishing
    use_vellum_emitter = is_events_emitting_enabled(executor_context)
    emitters: list["BaseWorkflowEmitter"] = []
    if use_vellum_emitter:
        emitters = [VellumEmitter()]

    use_vellum_resolver = executor_context.previous_execution_id is not None
    resolvers: list["BaseWorkflowResolver"] = []
    if use_vellum_resolver:
        resolvers = [VellumResolver()]

    # Explicit constructor call to satisfy typing
    workflow = Workflow(
        context=workflow_context,
        store=EmptyStore(),
        emitters=emitters,
        resolvers=resolvers,
    )

    return workflow, namespace


def _create_workflow_context(executor_context: BaseExecutorContext) -> WorkflowContext:
    if executor_context.environment_variables:
        os.environ.update(executor_context.environment_variables)

    namespace = _get_file_namespace(executor_context)

    return WorkflowContext(
        vellum_client=executor_context.vellum_client,
        execution_context=executor_context.execution_context,
        generated_files=executor_context.files,
        namespace=namespace,
    )


def _get_file_namespace(executor_context: BaseExecutorContext) -> str:
    if (
        LOCAL_WORKFLOW_MODULE
        and hasattr(executor_context.execution_context.parent_context, "deployment_name")
        and LOCAL_DEPLOYMENT == executor_context.execution_context.parent_context.deployment_name  # type: ignore
    ):
        return LOCAL_WORKFLOW_MODULE

    if executor_context.execution_id:
        return str(executor_context.execution_id)

    return get_random_namespace()


def get_random_namespace() -> str:
    return "workflow_tmp_" + "".join(random.choice(string.ascii_letters + string.digits) for i in range(14))


def _enrich_event(event: WorkflowEvent, executor_context: Optional[BaseExecutorContext] = None) -> WorkflowEvent:
    """
    Enrich an event with metadata based on the event type.

    For initiated events, include server and SDK versions.
    For fulfilled events with WORKFLOW_DEPLOYMENT parent, include memory usage.
    """
    metadata: Optional[dict] = None

    try:
        is_deployment = event.parent and event.parent.type in ("WORKFLOW_DEPLOYMENT", "EXTERNAL")

        if event.name == "workflow.execution.initiated" and is_deployment:
            metadata = {
                **get_version(),
            }

            memory_mb = get_memory_in_use_mb()
            if memory_mb is not None:
                metadata["memory_usage_mb"] = memory_mb

            if executor_context is not None:
                metadata["is_new_server"] = executor_context.is_new_server
        elif event.name == "workflow.execution.fulfilled" and is_deployment:
            metadata = {}
            memory_mb = get_memory_in_use_mb()
            if memory_mb is not None:
                metadata["memory_usage_mb"] = memory_mb
    except Exception:
        pass

    vellum_client = executor_context.vellum_client if executor_context else None
    return event_enricher(event, vellum_client, metadata=metadata)


def _dump_event(event: BaseEvent, executor_context: BaseExecutorContext) -> dict:
    module_base = executor_context.module.split(".")

    dump = event.model_dump(
        mode="json",
        context={"event_enricher": lambda event: _enrich_event(event, executor_context)},
    )
    if dump["name"] in {
        "workflow.execution.initiated",
        "workflow.execution.fulfilled",
        "workflow.execution.rejected",
        "workflow.execution.streaming",
        "workflow.execution.paused",
        "workflow.execution.resumed",
    }:
        dump["body"]["workflow_definition"]["module"] = module_base + dump["body"]["workflow_definition"]["module"][1:]
    elif dump["name"] in {
        "node.execution.initiated",
        "node.execution.fulfilled",
        "node.execution.rejected",
        "node.execution.streaming",
        "node.execution.paused",
        "node.execution.resumed",
    }:
        dump["body"]["node_definition"]["module"] = module_base + dump["body"]["node_definition"]["module"][1:]

    return dump


def _get_workflow_inputs(
    executor_context: BaseExecutorContext, workflow_class: Type[BaseWorkflow]
) -> Optional[BaseInputs]:
    if not executor_context.inputs:
        return None

    if not executor_context.files.get("inputs.py"):
        return None

    namespace = _get_file_namespace(executor_context)
    inputs_module_path = f"{namespace}.inputs"
    try:
        inputs_module = importlib.import_module(inputs_module_path)
    except Exception as e:
        raise WorkflowInitializationException(
            message=f"Failed to initialize workflow inputs: {e}",
            workflow_definition=workflow_class,
        ) from e

    if not hasattr(inputs_module, "Inputs"):
        raise WorkflowInitializationException(
            message=f"Inputs module {inputs_module_path} does not have a required Inputs class",
            workflow_definition=workflow_class,
        )

    if not issubclass(inputs_module.Inputs, BaseInputs):
        raise WorkflowInitializationException(
            message=f"""The class {inputs_module_path}.Inputs was expected to be a subclass of BaseInputs, \
but found {inputs_module.Inputs.__class__.__name__}""",
            workflow_definition=workflow_class,
        )

    return inputs_module.Inputs(**executor_context.inputs)
